/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2009-2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.talend;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.su.ServiceEngineServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;
import org.ow2.petals.se.talend.interim.WsdlMessageChecker;
import org.ow2.petals.se.talend.model.TalendJob;
import org.ow2.petals.se.talend.model.TalendJobHandler;

/**
 * The service-unit manager for the Talend service engine.
 * <p>
 * This manager gets and checks the provides block of the JBI descriptor.<br />
 * It also loads the Talend job class. If this class loading fails, the deployment fails.
 * </p>
 * 
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class TalendSuManager extends ServiceEngineServiceUnitManager {

    /**
     * The service engine instance.
     */
    private final TalendSe talendSe;

    private final Map<String,TalendJobHandler> suNameToJobHandlers =
        new ConcurrentHashMap<String,TalendJobHandler> ();


    /**
     * Constructor.
     * @param talendSe the service engine instance
     */
    public TalendSuManager( TalendSe talendSe ) {
        super( talendSe );
        this.talendSe = talendSe;
    }


    /**
     * Creates and stores a {@link TalendJob} instance describing the deployed job.
     * @see org.ow2.petals.component.framework.su.AbstractServiceUnitManager
     * #doDeploy(java.lang.String,java.lang.String,org.ow2.petals.component.framework.jbidescriptor.generated.Jbi)
     */
    @Override
    protected void doDeploy( String serviceUnitName, String suRootPath, Jbi jbi )
    throws PEtALSCDKException {

        // Check the JBI descriptor
        if( jbi == null
                || jbi.getServices() == null
                || jbi.getServices().getProvides() == null
                || jbi.getServices().getProvides().size() == 0 )
            throw new PEtALSCDKException( "Invalid JBI descriptor: the submitted one does not contain a 'provides' section." );

        if( jbi.getServices().getProvides().size() != 1 )
            throw new PEtALSCDKException( "Invalid JBI descriptor: the submitted one contains more than one 'provides' section." );

        Provides provides = jbi.getServices().getProvides().get( 0 );
        if( provides == null )
            throw new PEtALSCDKException( "Invalid JBI descriptor: the 'provides' section is invalid." );


        // Get the job parameters
        ServiceUnitDataHandler suDataHandler = getSUDataHandlerForProvides( provides );
        if( suDataHandler == null )
            throw new PEtALSCDKException( "Error while processing the JBI descriptor in the component. The SU data handler was null." );

        ConfigurationExtensions extensions = suDataHandler.getConfigurationExtensions( provides );
        if( extensions == null )
            throw new PEtALSCDKException( "Invalid JBI descriptor: the submitted one does not contain any component extension." );

        String name = extensions.get( "name" );
        if( name == null )
            name = "";

        String className = extensions.get( "class-name" );
        if( className == null || className.trim().length() == 0 )
            throw new PEtALSCDKException( "Invalid JBI descriptor: the class-name element has an empty value while a qualified class name was expected." );

        String context = extensions.get( "context" );
        if( context == null )
            context = "";

        String singletonAsString = extensions.get( "singleton" );
        boolean singleton = Boolean.parseBoolean( singletonAsString );
        String edptName = provides.getEndpointName();

        // Store the output attachments that must be set on the exchange after the job execution
        // Keep the LinkedHashMap (we want to conserve the order)
        List<String> outAttachmentContextNames = new ArrayList<String> ();
        for( Map.Entry<String,String> entry : extensions.entrySet()) {
            if( entry.getKey().startsWith( "output-attachment" ))
                outAttachmentContextNames.add( entry.getValue());
        }


        // Build and store the job
        TalendJob job = new TalendJob(
                name,
                className,
                context,
                singleton,
                outAttachmentContextNames,
                edptName,
                suRootPath,
                serviceUnitName );

        // TODO: delete this code portion as soon as WSDL-based validation has moved into the CDK
        // Add a WSDL-based validation delegate?
        TalendJobHandler talendJobHandler = new TalendJobHandler( job );
        String validateExchangeAsString = extensions.get( "validate-exchange-by-wsdl" );
        boolean validateExchangeBySu = Boolean.parseBoolean( validateExchangeAsString );
        boolean validateExchangeByComp = this.talendSe.validateExchangeByWsdl();

        if( validateExchangeBySu || validateExchangeByComp ) {

            String wsdlElementValue = provides.getWsdl();
            URL wsdlLocation = null;
            if( wsdlElementValue != null
                    && wsdlElementValue.trim().length() > 0 ) {

                try {
                    wsdlLocation = new URL( wsdlElementValue );

                } catch( final MalformedURLException e ) {

                    File wsdlFile = new File( suRootPath, wsdlElementValue );
                    try {
                        if( wsdlFile.exists())
                            wsdlLocation = wsdlFile.toURI().toURL();

                    } catch( final MalformedURLException e1 ) {
                        // nothing
                    }
                }
            }

            if( wsdlLocation != null ) {
                try {
                    WsdlMessageChecker msgChecker = new WsdlMessageChecker(
                            wsdlLocation, provides.getInterfaceName(), this.logger );

                    talendJobHandler.setWsdlMessageChecker( msgChecker );

                } catch( WSDLException e ) {
                    throw new PEtALSCDKException( "The provided WSDL does not match the component validation criteria." );
                }
            }
        }
        // End of code portion

        // Register the job handler
        this.talendSe.registerTalendJobHandler( talendJobHandler );
        this.suNameToJobHandlers.put( serviceUnitName, talendJobHandler );
    }


    /* (non-Javadoc)
     * @see org.ow2.petals.component.framework.su.AbstractServiceUnitManager
     * #doInit(java.lang.String, java.lang.String)
     */
    @Override
    protected void doInit( String serviceUnitName, String suRootPath ) throws PEtALSCDKException {

        try {
            TalendJobHandler jobHandler = this.suNameToJobHandlers.get( serviceUnitName );
            jobHandler.start();

        } catch( Exception e ) {
            throw new PEtALSCDKException( e );
        }
    }


    /* (non-Javadoc)
     * @see org.ow2.petals.component.framework.su.AbstractServiceUnitManager
     * #doUndeploy(java.lang.String)
     */
    @Override
    protected void doShutdown( String serviceUnitName ) throws PEtALSCDKException {

        try {
            TalendJobHandler jobHandler = this.suNameToJobHandlers.get( serviceUnitName );
            jobHandler.stop();

        } catch( Exception e ) {
            throw new PEtALSCDKException( e );
        }
    }


    /* (non-Javadoc)
     * @see org.ow2.petals.component.framework.su.AbstractServiceUnitManager
     * #doUndeploy(java.lang.String)
     */
    @Override
    protected void doUndeploy( String serviceUnitName ) throws PEtALSCDKException {

        try {
            TalendJobHandler jobHandler = this.suNameToJobHandlers.remove( serviceUnitName );
            jobHandler.releaseResources();
            this.talendSe.unregisterTalendJobHandler( jobHandler.getEndpointName());
            jobHandler = null;

        } catch( Exception e ) {
            throw new PEtALSCDKException( e );
        }
    }
}
