/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2009-2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.talend.model;

import java.util.List;

import org.ow2.petals.se.talend.interim.WsdlMessageChecker;
import org.ow2.petals.se.talend.utils.TalendClassLoaderUtils;
import org.ow2.petals.se.talend.utils.TalendURLClassLoader;


/**
 * A class that use {@link TalendJob} information to manage the life cycle of job instances.
 * <p>
 * This class acts as a proxy, so that no one access the internal model directly.
 * This class also provides utilities to determine what actions can be undertaken for a target job.
 * </p>
 * <p>
 * An instance of this class can only be created by the SU manager, when
 * a service-unit for the Talend SE is deployed. Undeployment of the same
 * service-unit should result in the destruction of this instance.
 * </p>
 * <p>
 * This class must only rely on the internal model of the component, not on
 * specific parameters of the SU.
 * </p>
 * 
 * @author Vincent Zurczak - EBM WebSourcing
 * 
 *  <p>
 * FIXME: should we create a second implementation using a pool of job instances?
 * <br />
 * Job class instantiation will be mandatory anyway, because we cannot make any
 * assumption on what code is executed by the job. There is no reset() method in jobs
 * that we might call.
 * </p>
 */
public class TalendJobHandler {

	/**
	 * The Talend job this instance manages.
	 */
	private final TalendJob talendJob;

	/**
	 * Computed field: true if one instance of this job is running, false otherwise.
	 */
	private boolean hasRunningInstance = false;

	/**
	 * An object in charge of validating exchanges with respect to the provided WSDL.
	 * <p>
	 * This object is intended to disappear in the next version.<br />
	 * WSDL-based validation should be moved in the CDK.
	 * </p>
	 */
	private WsdlMessageChecker wsdlMessageChecker;

	/**
	 * An object that will be used as a lock in the management of the instance's state.
	 */
	private final Object lock = new Object();

	/**
	 * Is the associated SU stopped?
	 */
	private boolean loaded = false;

	/**
	 * The job's class.
	 */
	private Class<?> jobClass;


	/**
	 * Constructor.
	 * @param talendJob
	 */
	public TalendJobHandler( TalendJob talendJob ) {
		this.talendJob = talendJob;
	}


	/**
	 * Starts the job handler.
	 * <p>
	 * Starting it
	 * </p>
	 * 
	 * @throws ClassNotFoundException if the class was not found in the SU class path
	 */
	public void start() throws ClassNotFoundException {

		synchronized( this.lock ) {

			// We do not want this class loader to inherit from the SE's class loader
			ClassLoader classLoader = TalendClassLoaderUtils.createClassLoader(
					this.talendJob.getSuRootPath(),
					ClassLoader.getSystemClassLoader());

			this.jobClass = classLoader.loadClass( this.talendJob.getClassName());
			this.loaded = true;
		}
	}


	/**
	 * Stops the job handler.
	 */
	public void stop() {

		synchronized( this.lock ) {

			ClassLoader cl = this.jobClass.getClassLoader();
			this.jobClass = null;
			this.loaded = false;
			if( cl instanceof TalendURLClassLoader ) {
				((TalendURLClassLoader) cl).close();
				cl = null;
			}
		}
	}


	/**
	 * Creates a new instance of the job.
	 * <p>
	 * Creating a new instance should only be done if the creation
	 * was granted by {@link #canCreateNewJobInstance()}.
	 * </p>
	 * 
	 * @return the new instance
	 * @throws InstantiationException if the instantiation failed
	 * @throws IllegalAccessException if the instantiation was not allowed
	 */
	public Object createNewJobInstance() throws InstantiationException, IllegalAccessException {
		return this.jobClass.newInstance();
	}


	/**
	 * Tries to get the permission to create a new instance.
	 * <p>
	 * If the job is singleton and this method returns true, then no other call
	 * to {@link #canCreateNewJobInstance()} will return true until this handler is
	 * released.
	 * </p>
	 * <p>
	 * If the job is not singleton, this method always return true and nothing is blocked.
	 * </p>
	 * <p>
	 * This method must always be called before {@link #createNewJobInstance()}.<br />
	 * Checking the running state is essential for singleton jobs, and without effect for the others.<br />
	 * A call to this method must always be followed by a call to {@link #releaseRunningState()}
	 * when the execution is over.
	 * </p>
	 * <p>
	 * If the job is singleton, then its returns true only if there is no running instance.<br />
	 * If the job is not singleton, then it always return true.
	 * </p>
	 * <p>
	 * Acquiring the running state is atomic.
	 * <p/>
	 * <p>
	 * The fact that a job instance may be (or cannot be) created is very important
	 * for the component (SE). The action to undertake in both cases depends on
	 * the component logic. This is why {@link #canCreateNewJobInstance()} and {@link #createNewJobInstance()}
	 * should not be merged.
	 * </p>
	 * 
	 * @return true if a new instance of this job can be created, false otherwise
	 */
	public boolean canCreateNewJobInstance() {

		boolean acquired = false;
		if( this.loaded ) {
			synchronized( this.lock ) {
				if( ! this.talendJob.isSingleton() || ! this.hasRunningInstance ) {
					acquired = true;
					this.hasRunningInstance = true;
				}
			}
		}

		return acquired;
	}


	/**
	 * Releases the running state of the job.
	 * <p>
	 * If a new job instance could not be created (because it was singleton and that it was
	 * already running), then calling this method will enable the creation of new job instances.
	 * </p>
	 * <p>
	 * Calling this method is essential for singleton jobs, and without effect for the others.<br />
	 * This method must be called as soon as a job instance has terminated.
	 * </p>
	 * <p>
	 * Releasing the running state (which means unlock the creation of new instance) is an
	 * atomic operation (synchronized).
	 * </p>
	 */
	public void releaseRunningState() {

		synchronized( this.lock ) {
			this.hasRunningInstance = false;
		}
	}


	/**
	 * @return the WSDL message checker (can be null)
	 * <p>
	 * If the result is null, then either the WSDL was invalid, or
	 * exchanges are not to be validated.
	 * </p>
	 */
	public WsdlMessageChecker getWsdlMessageChecker() {
		return this.wsdlMessageChecker;
	}


	/**
	 * @param wsdlMessageChecker the wsdlMessageChecker to set
	 */
	public void setWsdlMessageChecker( WsdlMessageChecker wsdlMessageChecker ) {
		this.wsdlMessageChecker = wsdlMessageChecker;
	}


	/**
	 * @return the end-point name associated with this handler
	 */
	public String getEndpointName() {
		return this.talendJob.getEndpointName();
	}


	/**
	 * @return the job context
	 */
	public String getContext() {
		return this.talendJob.getContext();
	}


	/**
	 * @return the job name
	 */
	public String getName() {
		return this.talendJob.getName();
	}

	/**
	 * @return a list of context names to retrieve
	 * @see org.ow2.petals.se.talend.model.TalendJob#getOutAttachmentContextNames()
	 */
	public List<String> getOutAttachmentContextNames() {
		return this.talendJob.getOutAttachmentContextNames();
	}


	/**
	 * Releases the resources.
	 */
	public void releaseResources() {

		if( this.wsdlMessageChecker != null )
			this.wsdlMessageChecker.deleteTamporaryFiles();
	}
}
