/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2009-2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.talend.processor;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.jbi.messaging.MessagingException;

import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.ow2.petals.se.talend.TalendSe;
import org.ow2.petals.se.talend.interim.WsdlMessageChecker;
import org.ow2.petals.se.talend.model.TalendBusinessException;
import org.ow2.petals.se.talend.model.TalendJobHandler;
import org.ow2.petals.se.talend.model.TalendTechnicalException;
import org.ow2.petals.se.talend.utils.FileContentDataSource;
import org.ow2.petals.se.talend.utils.TalendUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * A basis processor for exchanges addressed to the Petals-SE-Talend component.
 * @author Vincent Zurczak - EBM WebSourcing
 */
public abstract class AbstractExchangeProcessor {

	// All the WSDLs and SUs use the same name space.
	public static final String WSDL_TNS = "http://petals.ow2.org/talend/";
	public static final String JAXB_TNS = "http://jaxb.dev.java.net/array";
	public static final String MTOM_TNS = "http://www.w3.org/2004/08/xop/include";
	public static final String XSI_TNS = "http://www.w3.org/2001/XMLSchema-instance";

	protected String logHint;
	protected final Logger log;
	protected final TalendSe talendSe;


	/**
	 * Constructor.
	 * @param talendSe
	 */
	protected AbstractExchangeProcessor( TalendSe talendSe ) {

		assert( talendSe != null );
		this.talendSe = talendSe;
		this.log = talendSe.getLogger();
	}


	/**
	 * Called when the an exchange processing is delegated to the right MEP processor.
	 * <p>
	 * This method has three things to do on every call:
	 * </p>
	 * <ul>
	 *  <li>It must set {@link #logHint}.</li>
	 *  <li>It must call {@link #validateAndRun(Exchange)}.</li>
	 *  <li>It must handle the possible exceptions and update the exchange according to the MEP.</li>
	 * </ul>
	 * 
	 * @param logHint the log hint (identifies the exchange)
	 * @param exchange the exchange
	 * @return true to let the CDK send the response, false otherwise
	 */
	public abstract boolean onJbiMessage( String logHint, Exchange exchange );


	/**
	 * Instantiates, prepares the job parameters, executes the job and updates the result, if any.
	 * <p>
	 * There is no need the release the job's running state, as it is already done in
	 * {@link #validateAndRun(Exchange)}.
	 * </p>
	 * 
	 * @param talendJobHandler the Talend job handler
	 * @param exchange the exchange
	 * @throws TalendBusinessException if an exception occurred and is directly related to the job
	 * @throws TalendTechnicalException if an exception occurred and is related to the Petals context
	 */
	protected abstract void instantiateAndExecute( TalendJobHandler talendJobHandler, Exchange exchange )
	throws TalendBusinessException, TalendTechnicalException;


	/**
	 * Validates the exchange (according to the configuration) and starts the job execution if it is possible.
	 * <p>
	 * If the validation succeeded, {@link #instantiateAndExecute(TalendJobHandler, Exchange)} is called.
	 * It is up to this second method to manage the errors or faults to set on the exchange.
	 * Faults and errors must be set according to the MEP.
	 * </p>
	 * 
	 * @param exchange the exchange
	 * @throws TalendTechnicalException
	 * @throws TalendBusinessException
	 */
	protected void validateAndRun( Exchange exchange )
	throws TalendTechnicalException, TalendBusinessException {

		// Write down the input message (DEBUG)
		// Keep it, it is painful to write it from zero every time
		//        try {
		//            Source source = new DOMSource( exchange.getInMessageContentAsDocument( true ));
		//            File file = new File( "C:/Documents and Settings/vzurczak/Bureau/inMsg.xml" );
		//            Result result = new StreamResult( file );
		//            Transformer xformer = TransformerFactory.newInstance().newTransformer();
		//            xformer.transform( source, result );
		//
		//        } catch( Exception e1 ) {
		//            e1.printStackTrace();
		//        }

		// Keep the end-point name (throws exception)
		String edpt;
		try {
			edpt = exchange.getEndpointName();
		} catch( Exception e ) {
			throw new TalendTechnicalException( e.getMessage());
		}

		// Provider role only
		if( ! exchange.isProviderRole()) {
			String msg = "The provider role is the only role being supported in the Petals-SE-Talend component.";
			if( this.log.isLoggable( Level.INFO ))
				this.log.info( this.logHint + " encountered a problem. " + msg );

			throw new TalendTechnicalException( msg );
		}

		// Get the job
		TalendJobHandler talendJobHandler = this.talendSe.getTalendJobHandler( edpt );

		// FIXME: this code portion will have to be deleted once WSDL-based validation is in the CDK
		// Validate the exchange with respect to the WSDL?
		WsdlMessageChecker exchangeChecker = talendJobHandler.getWsdlMessageChecker();
		if( exchangeChecker != null  ) {
			if( ! exchangeChecker.validateExchange( exchange )) {
				String msg = "Invalid message exchange with respect to the service's WSDL.";
				if( this.log.isLoggable( Level.INFO ))
					this.log.info( this.logHint + " encountered a problem. " + msg );

				throw new TalendTechnicalException( msg );
			}
		}
		// End of code portion

		// Process the job and release it
		if( talendJobHandler.canCreateNewJobInstance()) {
			try {
				instantiateAndExecute( talendJobHandler, exchange );
			} finally {
				talendJobHandler.releaseRunningState();
			}
		}
		else {
			String msg = "A new job instance could not be created. The job must be singleton and already running.";
			if( this.log.isLoggable( Level.INFO ))
				this.log.info( this.logHint + " could not be processed. " + msg );

			throw new TalendBusinessException( msg );
		}
	}


	/**
	 * Gets the contexts from the input message.
	 * @param inputDocument the input message
	 * @return the list contexts for the job
	 */
	protected Map<String,String> getContexts( Document inputDocument ) {

		Map<String,String> contexts = new HashMap<String,String> ();
		NodeList contextsCandidatesList = inputDocument.getElementsByTagNameNS( "*", "contexts" );
		Node contextsRootElement = null;

		if( contextsCandidatesList.getLength() > 0 ) {
			contextsRootElement = contextsCandidatesList.item( 0 );
			NodeList contextsList = ((Element) contextsRootElement).getChildNodes();

			for( int j=0; j<contextsList.getLength(); j++ ) {
				Node contextElt = contextsList.item( j );
				if( contextElt.getNodeType() != Node.ELEMENT_NODE )
					continue;

				String name = TalendUtils.getNodeName( contextElt );
				String value = contextElt.getTextContent().trim();
				if( name != null && value.trim().length() > 0 )
					contexts.put( name, value );
			}
		}
		else if( this.log.isLoggable( Level.FINE )) {
			this.log.fine( this.logHint + ": no context was found in the input message." );
		}

		return contexts;
	}


	/**
	 * Gets the input data beans from the message.
	 * @param inputDocument the message
	 * @return a document which only contains the data beans to pass to a tPetalsInput instance
	 */
	protected Document getInputDataBeans( Document inputDocument ) {

		NodeList contentNodes = inputDocument.getElementsByTagNameNS( "*", "in-data-bean" );
		Document inputContentDocument = XMLUtil.createDocument();
		Element rootElement = inputContentDocument.createElement( "root" );
		inputContentDocument.appendChild( rootElement );

		for( int i=0; i<contentNodes.getLength(); i++ ) {
			Node n = inputContentDocument.importNode( contentNodes.item( i ), true );
			rootElement.appendChild( n );
		}

		if( contentNodes.getLength() == 0
				&& this.log.isLoggable( Level.FINE )) {
			this.log.fine( this.logHint + ": no in-data-bean was found in the input message." );
		}

		return inputContentDocument;
	}


	/**
	 * Adds the input attachments in the context map.
	 * @param inputDocument
	 * @param tempDir
	 * @param exchange
	 * @param contexts
	 * @throws TalendTechnicalException
	 */
	protected void addInputAttachments( Document inputDocument, File tempDir, Exchange exchange, Map<String,String> contexts )
	throws TalendTechnicalException {

		NodeList inAttachmentsCandidatesList = inputDocument.getElementsByTagNameNS( "*", "in-attachments" );
		Node inAttachmentsRootElement = null;

		if( inAttachmentsCandidatesList.getLength() > 0 ) {
			inAttachmentsRootElement = inAttachmentsCandidatesList.item( 0 );
			NodeList inAttachmentsList = ((Element) inAttachmentsRootElement).getChildNodes();

			for( int j=0; j<inAttachmentsList.getLength(); j++ ) {
				Node inAttachmentElt = inAttachmentsList.item( j );
				if( inAttachmentElt.getNodeType() == Node.ELEMENT_NODE )
					processInAttachment((Element) inAttachmentElt, exchange, tempDir, contexts );
			}
		}
		else if( this.log.isLoggable( Level.FINE ))
			this.log.fine( this.logHint + ": no input attachment was found in the input message." );
	}


	/**
	 * Parses the element to build a context based on the attachment type.
	 * <p>
	 * The element name is expected to be of the form "String|File_contextName".<br />
	 * String-based transfer mode means the attachment is set into a String and passed
	 * this way to the job. File-based transfer mode means the attachments is written into
	 * a temporary file, and the file location is passed as a context to the job.
	 * </p>
	 * 
	 * @param inAttachmentElt the element to process
	 * @param exchange
	 * @param tempDir
	 * @param contexts the map in which the contexts must be set
	 * @throws TalendTechnicalException
	 */
	protected void processInAttachment(
			Element inAttachmentElt,
			Exchange exchange,
			File tempDir,
			Map<String, String> contexts )
	throws TalendTechnicalException {

		String contextName = TalendUtils.getNodeName( inAttachmentElt );

		// SWA-like (supported to be less strict, but not supported for the returned attachments).
		// Real SWA needs a swaRef element, and the associated WSDL needs mime declarations.
		// It is not interesting to support it, and MTOM is the way W3C recommends.
		String attachmentId = null ;
		if( inAttachmentElt.getTextContent() != null )
			attachmentId = inAttachmentElt.getTextContent().trim().replaceFirst( "cid:", "" );

		// MTOM
		if( attachmentId == null || attachmentId.length() == 0 ) {
			NodeList includeNodes = inAttachmentElt.getElementsByTagNameNS( MTOM_TNS, "*" );
			for( int i=0; i<includeNodes.getLength(); i++ ) {

				Node n = includeNodes.item( i );
				if( "include".equalsIgnoreCase( TalendUtils.getNodeName( n ))) {
					attachmentId = ((Element) n).getAttribute( "href" ).replaceFirst( "cid:", "" );
					break;
				}
			}
		}

		if( attachmentId == null || attachmentId.length() == 0 ) {
			String msg = "The attachment ID for " + contextName + " could not be determined.";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendTechnicalException( msg );
		}

		try {
			DataHandler dataHandler = null;
			for( String attName : exchange.getInMessageAttachmentNames()) {
				if( TalendUtils.compare( attachmentId, attName )) {
					dataHandler = exchange.getInMessageAttachment( attName );
					break;
				}
			}

			if( dataHandler == null ) {
				String msg = "The attachment for " + contextName + " was not found in the input message.";
				if( this.log.isLoggable( Level.FINE ))
					this.log.fine( this.logHint + " encountered a problem. " + msg );

				throw new TalendTechnicalException( msg );
			}

			// File: write the attachment into a file (yes, it takes time...)
			if( ! tempDir.exists())
				createDirectory( tempDir );

			String fileName = dataHandler.getName();
			if( fileName == null )
				fileName = "";

			// DataHandlers can return a full file location instead of just a file name
			String f = fileName.replaceAll( "\\\\", "/" );
			int index = f.lastIndexOf( '/' );
			if( index > 0 )
				fileName = f.substring( index );

			if( fileName.trim().length() == 0 )
				fileName = UUID.randomUUID().toString();

			File attachmentFile = new File( tempDir, fileName );

			// Here, we will have performance loss
			TalendUtils.copyStream( dataHandler.getInputStream(), attachmentFile );
			String contextValue = attachmentFile.getAbsolutePath();
			contexts.put( contextName, contextValue );

		} catch( MessagingException e ) {
			String msg = "The attachment for " + contextName + " could not be read.";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendTechnicalException( msg );

		} catch( IOException e ) {
			String msg = "The attachment for " + contextName + " could not be processed.";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendTechnicalException( msg );
		}
	}


	/**
	 * Gets the Talend options from the input message.
	 * @param inputDocument the input message
	 * @return a list (potentially empty) of Talend options
	 */
	protected List<String> getTalendOptions( Document inputDocument ) {

		NodeList talendOptionsList = inputDocument.getElementsByTagNameNS( "*", "talend-option" );
		List<String> talendOptions = new ArrayList<String> ();
		for( int i=0; i<talendOptionsList.getLength(); i++ ) {
			String option = talendOptionsList.item( i ).getTextContent();
			talendOptions.add( option.trim());
		}

		if( talendOptionsList.getLength() == 0
				&& this.log.isLoggable( Level.FINE )) {
			this.log.fine( this.logHint + ": no Talend option was found in the input message." );
		}

		return talendOptions;
	}


	/**
	 * Initializes the job instance and executes it.
	 * @param instance the job instance
	 * @param talendJobHandler the job handler
	 * @param contexts the overridden contexts
	 * @param inputDataBeansDocument the input data beans (tPetalsInput)
	 * @param talendOptions the talend options
	 * @return the job's execution result
	 * @throws TalendBusinessException if the run method does not exist
	 */
	protected String[][] executeJob(
			Object instance,
			TalendJobHandler talendJobHandler,
			Map<String,String> contexts,
			Document inputDataBeansDocument,
			List<String> talendOptions )
	throws TalendBusinessException {

		Method runMethod;
		try {
			runMethod = instance.getClass().getMethod( "runJob", String[].class );

		} catch( NoSuchMethodException e ) {
			String msg = "The run operation was not available.";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendBusinessException( msg );
		}

		List<String> args = new ArrayList<String> ();

		// Prepare the contexts (including input attachments)
		args.add( "--context=" + talendJobHandler.getContext());
		for( Map.Entry<String,String> entry : contexts.entrySet()) {
			args.add( "--context_param" );
			args.add( entry.getKey() + "=" + entry.getValue());
		}

		// Prepare the other Talend parameters
		args.addAll( talendOptions );
		String[] argsArray = new String[ args.size()];
		args.toArray( argsArray );

		try {
			// Set the input data beans (tPetalsInput)
			try {
				Method setInMethod = instance.getClass().getMethod( "setPetalsInMessage", Document.class );
				if( setInMethod != null ) {
					setInMethod.invoke( instance, new Object[] { inputDataBeansDocument });
					setInMethod = null;
				}

			} catch( NoSuchMethodException e ) {

				// Maybe not an important error: log it and go on
				if( this.log.isLoggable( Level.FINE ))
					this.log.fine( this.logHint + ": input data beans could not be passed to the job (operation not supported)." );

				// Did the message have data beans to pass to the job?
				NodeList children = inputDataBeansDocument.getElementsByTagNameNS( "*", "in-data-bean" );
				if( children.getLength() > 1 ) {
					String msg = "Input data beans were found but could not be passed to the job (operation not supported).";
					if( this.log.isLoggable( Level.INFO ))
						this.log.info( this.logHint + " encountered a problem. " + msg );

					throw new TalendBusinessException( msg );
				}
			}

			// Execute the job
			String[][] jobResult = (String[][]) runMethod.invoke( instance, new Object[] { argsArray });
			runMethod = null;
			args = null;
			return jobResult;

		} catch( SecurityException e ) {
			String msg = "The job could not be executed (SecurityException).";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendBusinessException( msg );

		} catch( IllegalArgumentException e ) {
			String msg = "The job could not be executed (IllegalArgumentException).";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendBusinessException( msg );

		} catch( IllegalAccessException e ) {
			String msg = "The job could not be executed (IllegalAccessException).";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendBusinessException( msg );

		} catch( InvocationTargetException e ) {
			String msg = "The job could not be executed (InvocationTargetException).";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendBusinessException( msg );
		}

		// Note: these exceptions are considered business and not technical exceptions,
		// because they are related to the job's class.
	}


	/**
	 * Gets the output data beans from the job.
	 * @param instance the job instance
	 * @return a list of Element, each element being a row of data
	 * @throws TalendBusinessException if something went wrong
	 */
	protected List<Element> getOutputDataBeans( Object instance )
	throws TalendBusinessException {

		List<Element> result = new ArrayList<Element> ();
		try {
			Method setOutMethod = instance.getClass().getMethod( "getPetalsOutMessage" );
			Document outputContentDocument = null;
			if( setOutMethod != null ) {
				outputContentDocument = (Document) setOutMethod.invoke( instance, new Object[ 0 ]);
				setOutMethod = null;
			}

			if( outputContentDocument != null ) {
				NodeList beanNodes = outputContentDocument.getElementsByTagNameNS( "*", "outDataBean" );
				for( int i=0; i<beanNodes.getLength(); i++ )
					result.add((Element) beanNodes.item( i ));
			}

		} catch( NoSuchMethodException e ) {

			// Not an important error: log it and go on
			if( this.log.isLoggable( Level.INFO ))
				this.log.info( this.logHint + ": output data beans could not be retrieved (operation not supported)." );

		} catch( IllegalArgumentException e ) {
			String msg = "The output data beans could not be retrieved (IllegalArgumentException).";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendBusinessException( msg );

		} catch( IllegalAccessException e ) {
			String msg = "The output data beans could not be retrieved (IllegalAccessException).";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendBusinessException( msg );

		} catch( InvocationTargetException e ) {
			String msg = "The output data beans could not be retrieved (InvocationTargetException).";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendBusinessException( msg );
		}

		return result;
	}


	/**
	 * Sets the output attachments on the exchange.
	 * @param instance the job instance
	 * @param exchange the exchange
	 * @param talendJobHandler the Talend job handler
	 * @return a map (key = element name for the attachment, value = attachment ID)
	 * @throws TalendBusinessException if something went wrong
	 * @throws TalendTechnicalException
	 */
	protected Map<String,String> setOutputAttachments( Object instance, Exchange exchange, TalendJobHandler talendJobHandler )
	throws TalendBusinessException, TalendTechnicalException {

		// Keep the LinkedHashMap (we want to conserve the order)
		Map<String,String> attElementNameToAttId = new LinkedHashMap<String,String> ();
		List<String> cnat = talendJobHandler.getOutAttachmentContextNames();
		if( cnat.size() > 0 ) {
			try {

				Method getOutContextsMethod = null;
				try {
					getOutContextsMethod = instance.getClass().getMethod( "getContext" );

				} catch( NoSuchMethodException e ) {

					// Be here means output attachments were expected but cannot be retrieved.
					String msg = "Output attachments were expected but could not be retrieved (operation not supported).";
					if( this.log.isLoggable( Level.INFO ))
						this.log.info( this.logHint + " encountered a problem. " + msg );

					throw new TalendBusinessException( msg );
				}

				Properties contexts = null;
				if( getOutContextsMethod != null ) {
					contexts = (Properties) getOutContextsMethod.invoke( instance );
					getOutContextsMethod = null;
				}

				if( contexts == null )
					contexts = new Properties();

				for( String contextName : cnat ) {
					String attachmentContent = contexts.getProperty( contextName, null );
					try {
						if( attachmentContent == null ) {
							String msg = "The output attachment " + contextName + " was not found in the job (unexpected value).";
							if( this.log.isLoggable( Level.FINE ))
								this.log.fine( this.logHint + ": " + msg );

							throw new IOException( msg );
						}

						File attachmentFile = new File( attachmentContent );

						// One solution could be to use a FileDataSource.
						// But the file would have to remain on the disk.
						// After 300 executions, either we wrote in the same file 300 times, or we created 300 files.

						// Besides, if the temporary file was specified outside, an malicious client would be
						// able to crash Petals, by asking to create and use a new file on each call.

						// For safety reasons, we have to delete the file used for the output attachment.
						// We cannot trust the client.
						// This is why we have to load the entire file and pass it into a ByteArrayDataSource.

						DataSource dataSource = new FileContentDataSource( attachmentFile );
						DataHandler dataHandler = new DataHandler( dataSource );
						exchange.addOutMessageAttachment( contextName, dataHandler );
						attElementNameToAttId.put( contextName, contextName );

						// Delete the file
						if( ! attachmentFile.delete()) {
							String msg = "An attachment file could not be deleted (" + attachmentFile.getAbsolutePath() + ").";
							if( this.log.isLoggable( Level.FINE ))
								this.log.fine( this.logHint + " encountered a problem. " + msg );

							attachmentFile.deleteOnExit();
						}

					} catch( MessagingException e ) {
						String msg = "A file attachment could not be set on the exchange.";
						if( this.log.isLoggable( Level.FINE ))
							this.log.fine( this.logHint + " encountered a problem. " + msg );

						throw new TalendTechnicalException( msg );

					} catch( IOException e ) {
						String msg = "A file attachment was expected for " + contextName + " but could not be found.";
						if( this.log.isLoggable( Level.FINE ))
							this.log.fine( this.logHint + " encountered a problem. " + msg );

						throw new TalendBusinessException( msg );
					}
				}

			} catch( IllegalArgumentException e ) {
				String msg = "The output attachments could not be retrieved (IllegalArgumentException).";
				if( this.log.isLoggable( Level.FINE ))
					this.log.fine( this.logHint + " encountered a problem. " + msg );

				throw new TalendBusinessException( msg );

			} catch( IllegalAccessException e ) {
				String msg = "The output attachments could not be retrieved (IllegalAccessException).";
				if( this.log.isLoggable( Level.FINE ))
					this.log.fine( this.logHint + " encountered a problem. " + msg );

				throw new TalendBusinessException( msg );

			} catch( InvocationTargetException e ) {
				String msg = "The output attachments could not be retrieved (InvocationTargetException).";
				if( this.log.isLoggable( Level.FINE ))
					this.log.fine( this.logHint + " encountered a problem. " + msg );

				throw new TalendBusinessException( msg );
			}
		}

		return attElementNameToAttId;
	}


	/**
	 * Builds the output document that will be set on the exchange.
	 * @param jobResult the job's execution result
	 * @param outputDataBeans the output data beans (which come from a tPetalsOutput instance)
	 * @param attachmentNameToIds a map of elements names (for attachments) associated with an attachment ID
	 * @return the document to set on the exchange
	 */
	protected Document getOutputMessage( String[][] jobResult, List<Element> outputDataBeans, Map<String,String> attachmentNameToIds ) {

		Document outputDocument = XMLUtil.createDocument();
		Element rootElement = outputDocument.createElementNS( WSDL_TNS, "tns:executeJobResponse" );
		outputDocument.appendChild( rootElement );
		rootElement.setAttribute( "xmlns:xsi", XSI_TNS );
		rootElement.setAttribute( "xmlns:jaxb", JAXB_TNS );

		Element jobOutputElement = outputDocument.createElementNS( WSDL_TNS, "tns:talend-job-output" );
		rootElement.appendChild( jobOutputElement );

		// The job result
		for( String[] line : jobResult ) {
			Element executionResultElement = outputDocument.createElementNS( WSDL_TNS, "tns:executionResult" );
			jobOutputElement.appendChild( executionResultElement );

			for( String string : line ) {
				Element arrayElement = outputDocument.createElementNS( JAXB_TNS, "jaxb:item" );
				arrayElement.setTextContent( string );
				executionResultElement.appendChild( arrayElement );
			}
		}

		// Output attachments
		Element outAttachmentElement = outputDocument.createElementNS( WSDL_TNS, "tns:outAttachment" );
		jobOutputElement.appendChild( outAttachmentElement );
		for( Map.Entry<String,String> entry : attachmentNameToIds.entrySet()) {

			// Note: we did not look into the WSDL to determine which attachments are needed and in which order.
			// But the WSDL and the jbi.xml have been generated by the tools at the same moment.
			// And we use LinkedHashMaps to conserve the order defined during the export.
			// Therefore, we just assume this order has been preserved until here.
			Element attElement = outputDocument.createElementNS( WSDL_TNS, "tns:" + entry.getKey());
			outAttachmentElement.appendChild( attElement );

			// Attachments are returned in MTOM mode
			Element fileContentElt = outputDocument.createElementNS( WSDL_TNS, "tns:fileContent" );
			attElement.appendChild( fileContentElt );

			Element includeElt = outputDocument.createElementNS( MTOM_TNS, "xop:include" );
			includeElt.setAttribute( "href", "cid:" + entry.getValue());
			fileContentElt.appendChild( includeElt );
		}

		// Output data beans
		for( Element outputDataBeanElement : outputDataBeans ) {
			Node newNode = outputDocument.importNode( outputDataBeanElement, true );
			jobOutputElement.appendChild( newNode );
		}

		if( outputDataBeans.size() == 0
				&& this.log.isLoggable( Level.FINE )) {
			this.log.fine( this.logHint + ": no output data bean was found in the job." );
		}

		// Write down the output message (DEBUG)
		// Keep it, it is painful to write it from zero every time
		//		try {
		//			Source source = new DOMSource( outputDocument );
		//			File file = new File( "C:/Documents and Settings/vzurczak/Bureau/outMsg.xml" );
		//			Result result = new StreamResult( file );
		//			Transformer xformer = TransformerFactory.newInstance().newTransformer();
		//			xformer.transform( source, result );
		//
		//		} catch( Exception e1 ) {
		//			e1.printStackTrace();
		//		}

		return outputDocument;
	}


	/**
	 * Creates a directory in the scope of the Talend SE.
	 * @param tempDir
	 * @throws TalendTechnicalException if the file could not be created
	 */
	private void createDirectory( File tempDir ) throws TalendTechnicalException {

		if( ! tempDir.exists() && ! tempDir.mkdir()) {
			String msg = "Could not create a temporary directory to store attachments.";
			if( this.log.isLoggable( Level.FINE ))
				this.log.fine( this.logHint + " encountered a problem. " + msg );

			throw new TalendTechnicalException( msg );
		}
	}
}
