/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.talend.processor;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.logging.Level;
import javax.jbi.messaging.MessagingException;

import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.se.talend.TalendSe;
import org.ow2.petals.se.talend.model.TalendBusinessException;
import org.ow2.petals.se.talend.model.TalendJobHandler;
import org.ow2.petals.se.talend.model.TalendTechnicalException;
import org.ow2.petals.se.talend.utils.TalendUtils;
import org.w3c.dom.Document;

/**
 * The exchange processor for RobustInOnly MEP.
 * <p>
 * The only difference with the InOut processor is that we do not the
 * the result from the job, and that there is no possible output, except faults.
 * </p>
 * 
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class RobustInOnlyProcessor extends InOutProcessor {

    /**
     * Constructor.
     * @param talendSe
     */
    public RobustInOnlyProcessor( TalendSe talendSe ) {
        super( talendSe );
    }


    /* (non-Javadoc)
     * @see org.ow2.petals.se.talend.processor.AbstractExchangeProcessor
     * #instantiateAndExecute(
     * org.ow2.petals.se.talend.model.TalendJobHandler,
     * org.ow2.petals.component.framework.api.message.Exchange)
     */
    @Override
    public void instantiateAndExecute( TalendJobHandler talendJobHandler, Exchange exchange )
    throws TalendBusinessException, TalendTechnicalException {

        // Follow the execution flow
        if( this.log.isLoggable( Level.INFO ))
            this.log.info( this.logHint + ": request was validated, the job is about to be launched." );

        // Prepare a temporary directory for file attachment
        File tempDir = new File(
                System.getProperty( "java.io.tmpdir" ),
                UUID.randomUUID().toString());

        // From here, we need to go through the finally block.
        // Move code with caution...
        try {
            // Extract the data from the exchange
            Document inputDocument;
            try {
                inputDocument = exchange.getInMessageContentAsDocument();

            } catch( MessagingException e3 ) {
                String msg = "The input message could not be read.";
                if( this.log.isLoggable( Level.FINE ))
                    this.log.fine( this.logHint + " encountered a problem. " + msg );

                throw new TalendBusinessException( msg );
            }

            //
            // Prepare the input (4 possible input parts that can be mixed)
            // (a) Contexts
            Map<String,String> contexts = getContexts( inputDocument );

            // (b) Input data beans (for a tPetalsInput)
            Document inputDataBeansDocument = getInputDataBeans( inputDocument );

            // (c) Input attachment (that will be passed as contexts)
            addInputAttachments( inputDocument, tempDir, exchange, contexts );

            // (d) Other Talend options (that we do not add in documentation).
            // They are supported just in case where a Talend expert needs them.
            List<String> talendOptions = getTalendOptions( inputDocument );


            //
            // Create or get a job instance
            Object instance;
            try {
                instance = talendJobHandler.createNewJobInstance();

            } catch( InstantiationException e ) {
                String msg = "The job could not be instantiated (InstantiationException.";
                if( this.log.isLoggable( Level.FINE ))
                    this.log.fine( this.logHint + " encountered a problem. " + msg );

                throw new TalendBusinessException( msg );

            } catch( IllegalArgumentException e) {
                String msg = "The job could not be instantiated (IllegalArgumentException).";
                if( this.log.isLoggable( Level.FINE ))
                    this.log.fine( this.logHint + " encountered a problem. " + msg );

                throw new TalendBusinessException( msg );

            } catch( IllegalAccessException e ) {
                String msg = "The job could not be instantiated (IllegalAccessException).";
                if( this.log.isLoggable( Level.FINE ))
                    this.log.fine( this.logHint + " encountered a problem. " + msg );

                throw new TalendBusinessException( msg );
            }

            // And run it - we do not care about the result
            executeJob( instance, talendJobHandler, contexts, inputDataBeansDocument, talendOptions );
            if( this.log.isLoggable( Level.INFO ))
                this.log.info( this.logHint + " was successfully processed." );

            // Speed up the garbage collector's job
            contexts = null;
            inputDataBeansDocument = null;
            talendOptions = null;
            instance = null;

        } finally {

            // Delete the temporary directory, and all its content
            try {
                TalendUtils.deleteFilesRecursively( tempDir );
            } catch( IOException e ) {
                if( this.log.isLoggable( Level.FINE ))
                    this.log.fine(
                            this.logHint + " encountered a minor error. "
                            + tempDir.getAbsolutePath() + " could not be deleted." );
            }
        }
    }
}
